.toolbox {
  background: @theme_base_color;
  border: 1px solid @borders;
  border-bottom-left-radius: 9px;
  border-bottom-right-radius: 9px;
  border-top: 0;
  padding: 6px;
}

.mistakes {
  border-left: 1px solid @borders;
  border-right: 1px solid @borders;
}

.pronunciation {
  border: 1px solid @borders;
  border-top: 0;
}

.translation-frame  {
  background: @theme_base_color;
  border-top-left-radius: 9px;
  border-top-right-radius: 9px;
}
.translation-frame:disabled  {
  background: @insensitive_bg_color;
}
.translation-frame:backdrop  {
  background: @theme_unfocused_base_color;
}

.translation-frame scrollbar {
  padding-top: 3px;
}

.translation-frame scrollbar:dir(ltr) {
  border-top-right-radius: 9px;
}

.translation-frame scrollbar:dir(rtl) {
  border-top-left-radius: 9px;
}

.langselector {
  border-radius: 4px;
}

.bottomlang {
  background: @theme_base_color;
  border: 1px solid @borders;
  border-bottom: 0;
  border-left: 0;
  border-right: 0;
}
