# Copyright 2020 gi-lom
# Copyright 2020-2021 Mufeed Ali
# Copyright 2020-2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later


APP_ID = 'com.github.gi_lom.dialect'
RES_PATH = '/com/github/gi_lom/dialect'
MAX_LENGTH = 5000  # maximum number of characters you can translate at once
TRANS_NUMBER = 10  # number of translations to save in history
